----------------------------------------------------------------------------------------------------------------------------------------------------
-- WhatRepRecipes - View vendor sell values for all crafted items and reputation requirements, source, price and mats for all unskilled recipe ranks
----------------------------------------------------------------------------------------------------------------------------------------------------
-- Modules\Item.lua - Item and Recipes Stuff
--
-- Author: Expelliarm5s / June 2022 / All Rights Reserved
--
-- Version 1.2.10
-------------------------------------------------------------------------------------------------------
-- luacheck: ignore 212 globals DLAPI

-- luacheck: globals C_TradeSkillUI
-- luacheck: globals TradeSkillFrame
-- luacheck: globals RANK SPELL_REAGENTS AUCTION_CATEGORY_RECIPES SELL_PRICE ITEM_UNSELLABLE SPELL_REAGENTS_OPTIONAL
-- luacheck: globals GetDetailedItemLevelInfo

local addonName, addon = ...
local Item = addon:NewModule("Item", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName)
local private = {}
-------------------------------------------------------------------------------------------------------

------------------------------------------------------------------------------
-- Debug Stuff

function Item:DebugPrintf(...)
	if addon.isDebug then
		local status, res = pcall(format, ...)
		if status then
			addon:DebugLog("Itm~" .. res)
		end
	end
end

------------------------------------------------------------------------------
-- Login

function Item:Login()
	Item:DebugPrintf("Login()")
end

------------------------------------------------------------------------------
-- Recipe Infos for the recipe details frame

-- Return the recipe item link
function Item:GetItemLinkText(recipeID, rank)
	Item:DebugPrintf("GetItemLinkText(%s, %s)", tostring(recipeID), tostring(rank))

	local sourceText = ""
	if (rank) then
		sourceText = sourceText .. "|n|n|cFFFFD200" .. RANK .. " " .. tostring(rank) .. "|r: "
	else
		sourceText = sourceText .. "|n|n"
	end

	local minMade, maxMade = C_TradeSkillUI.GetRecipeNumItemsProduced(recipeID)
	if maxMade > 1 then
		if minMade == maxMade then
			sourceText = sourceText .. format("%s x ", tostring(minMade))
		else
			sourceText = sourceText .. format("%s-%s x ", tostring(minMade), tostring(maxMade))
		end
	end
	sourceText = sourceText .. (C_TradeSkillUI.GetRecipeItemLink(recipeID) or "")

	return sourceText
end

-- Return the recipe info text
function Item:GetRecipeInfo(recipeID, rank)
	Item:DebugPrintf("GetRecipeInfo(%s, %s)", tostring(recipeID), tostring(rank))

	local sourceText = Item:GetItemLinkText(recipeID, rank)

	sourceText = sourceText .. "|n" .. (C_TradeSkillUI.GetRecipeSourceText(recipeID) or "")
	if addon.isDebug then
		sourceText = sourceText .. "|n|cFFFFD200" .. AUCTION_CATEGORY_RECIPES .. " ID:|r " .. tostring(recipeID)
	end

	local numReagents = C_TradeSkillUI.GetRecipeNumReagents(recipeID)

	if numReagents > 0 then
		sourceText = sourceText .. "|n|cFFFFD200" .. SPELL_REAGENTS .. "|r"
		for reagentIndex = 1, numReagents do
			local reagentName, _, reagentCount = C_TradeSkillUI.GetRecipeReagentInfo(recipeID, reagentIndex);
			if reagentName then
				if reagentIndex > 1 then
					sourceText = sourceText .. "|n"
				end
				sourceText = sourceText .. tostring(reagentCount) .. " x " .. tostring(reagentName)
			end
		end
	end

	return sourceText
end

-- Return total and current recipe rank
function Item:CalculateRecipeRanks(recipeID)
	Item:DebugPrintf("CalculateRecipeRanks(%s)", tostring(recipeID))

	local totalRanks = 0;
	local currentRank = 0;
	local previousRecipeID = recipeID
	local recipeInfo

	while previousRecipeID do
		Item:DebugPrintf("  search for parent rank of %s", tostring(previousRecipeID))
		recipeInfo = C_TradeSkillUI.GetRecipeInfo(previousRecipeID)
		if recipeInfo then
			previousRecipeID = recipeInfo.previousRecipeID
		else
			previousRecipeID = nil
		end
	end

	if recipeInfo then
		local nextRecipeID = recipeInfo.recipeID
		while nextRecipeID do
			Item:DebugPrintf("  search for next rank of %s", tostring(nextRecipeID))
			recipeInfo = C_TradeSkillUI.GetRecipeInfo(nextRecipeID)
			if recipeInfo then
				totalRanks = totalRanks + 1
				if recipeID == nextRecipeID then
					currentRank = totalRanks
				end
				nextRecipeID = recipeInfo.nextRecipeID
			else
				nextRecipeID = nil
			end
		end
	end

	Item:DebugPrintf("  ... current=%s, total=%s", tostring(currentRank), tostring(totalRanks))

	return totalRanks, currentRank;
end

-- Return recipe and VendorSell infos
function Item:GetVendorSellInfo(recipeID)
	Item:DebugPrintf("GetVendorSellInfo(%s)", tostring(recipeID))

	local sourceText = ""
	sourceText = sourceText .. "|n|cFFFFD200" .. SELL_PRICE .. ":|r "

	local minMade, maxMade = C_TradeSkillUI.GetRecipeNumItemsProduced(recipeID)
	Item:DebugPrintf(" %s-%s per craft", tostring(minMade), tostring(maxMade))

	local link = C_TradeSkillUI.GetRecipeItemLink(recipeID)
	Item:DebugPrintf(" %s", tostring(link))

	local name, _, vendorSell = Item:GetItemInfo(link)

	Item:DebugPrintf(" %s / %s (%s) / %s", tostring(name), tostring(link), Item:EscapedLink(link), tostring(vendorSell))

	if name then
		if vendorSell == 0 then
			sourceText = sourceText .. ITEM_UNSELLABLE
		else
			if GetMoneyString and type(vendorSell) == "number" then
				sourceText = sourceText .. GetMoneyString(vendorSell)

				if C_TradeSkillUI.GetOptionalReagentInfo and C_TradeSkillUI.GetOptionalReagentInfo(recipeID)
					and C_TradeSkillUI.GetOptionalReagentInfo(recipeID)[1] then
					Item:DebugPrintf(" has optional reagents")
					sourceText = sourceText .. "|n|n|cFFFFD200" .. SELL_PRICE .. "/" .. SPELL_REAGENTS_OPTIONAL .. "|r"

					for _, v in ipairs(C_TradeSkillUI.GetOptionalReagentInfo(recipeID)) do
						local optionalReagentSlots = v.options

						-- all base item links of recipes with optional reagents obviously have the bonidID 3524
						-- just add the well-known bonusID of relics to determine the sell value
						-- should work somehow
						for i, id in ipairs(optionalReagentSlots) do
							local itemName = GetItemInfo(id)
							itemName = itemName or ("Relic-ItemID " .. tostring(id))
							Item:DebugPrintf(" i=%s, name=%s", tostring(i), tostring(itemName))
							local llink = Item:MakeRelicLink(link, id)
							if llink then
								local nname, lllink, nvendorSell = Item:GetItemInfo(llink)
								Item:DebugPrintf(" %s / %s (%s) / %s",
									tostring(nname), tostring(lllink), Item:EscapedLink(lllink), tostring(nvendorSell))
								if nname then
									sourceText = sourceText .. "|n" .. lllink .. "|n" .. GetMoneyString(nvendorSell) .. " (" .. itemName .. ")"
								end
							end
						end
					end
				end
			else
				sourceText = sourceText .. tostring(vendorSell)
			end
		end
	else
		sourceText = ""
	end

	return sourceText
end

-- Return new recipe info
function Item:GetRecipeSourceText(recipeInfo)
	Item:DebugPrintf("OK~GetRecipeSourceText: RecipeID %s", recipeInfo.recipeID)

	local sourceText = ""

	local totalRanks, currentRank = Item:CalculateRecipeRanks(recipeInfo.recipeID)

	Item:DebugPrintf(" currentRank=%s totalRanks=%s", tostring(currentRank), tostring(totalRanks))

	if totalRanks and tonumber(totalRanks) and tonumber(totalRanks) > 1 then
		local rank = currentRank

		-- check ranks
		while recipeInfo do
			Item:DebugPrintf(" try ID %s, learned=%s", tostring(recipeInfo.recipeID), tostring(recipeInfo.learned))
			if not recipeInfo.learned then
				sourceText = sourceText .. Item:GetRecipeInfo(recipeInfo.recipeID, rank)
				sourceText = sourceText .. Item:GetVendorSellInfo(recipeInfo.recipeID)
			else
				local vendorSellText = Item:GetVendorSellInfo(recipeInfo.recipeID)
				if vendorSellText and vendorSellText ~= "" then
					sourceText = sourceText .. Item:GetItemLinkText(recipeInfo.recipeID, rank)
					sourceText = sourceText .. vendorSellText
				end
			end
			if recipeInfo.nextRecipeID then
				recipeInfo = C_TradeSkillUI.GetRecipeInfo(recipeInfo.nextRecipeID)
			else
				recipeInfo = nil
			end
			rank = rank + 1
		end
	else
		local vendorSellText = Item:GetVendorSellInfo(recipeInfo.recipeID)
		if vendorSellText and vendorSellText ~= "" then
			sourceText = sourceText .. Item:GetItemLinkText(recipeInfo.recipeID)
			sourceText = sourceText .. vendorSellText
		end
	end

	return sourceText
end

-----------------------------------------------------------------------------
-- Helper Functions

-----------------------------------------------------------------------------
-- Work In Progress
private.vendorMats = {
			[124436] = {
				["copper"] = 40000,
				["link"] = "|cffffffff|Hitem:124436::::::::50:66:::::::|h[Fuchsblumenfluxus]|h|r",
				["ID"] = 124436,
			},
			[45087] = {
				["copper"] = 950000,
				["link"] = "|cff0070dd|Hitem:45087::::::::50:66:::::::|h[Runenbeschriebene Kugel]|h|r",
				["ID"] = 45087,
			},
			[160298] = {
				["copper"] = 3000,
				["link"] = "|cffffffff|Hitem:160298::::::::50:66:::::::|h[Langlebiger Fluxus]|h|r",
				["ID"] = 160298,
			},
			[2880] = {
				["copper"] = 100,
				["link"] = "|cffffffff|Hitem:2880::::::::50:66:::::::|h[Schwacher Fluxus]|h|r",
				["ID"] = 2880,
			},
			[2605] = {
				["copper"] = 100,
				["link"] = "|cffffffff|Hitem:2605::::::::50:66:::::::|h[Grüner Farbstoff]|h|r",
				["ID"] = 2605,
			},
			[47556] = {
				["copper"] = 1187500,
				["link"] = "|cff0070dd|Hitem:47556::::::::50:66:::::::|h[Kugel des Kreuzfahrers]|h|r",
				["ID"] = 47556,
			},
			[3466] = {
				["copper"] = 2000,
				["link"] = "|cffffffff|Hitem:3466::::::::50:66:::::::|h[Starker Fluxus]|h|r",
				["ID"] = 3466,
			},
			[49908] = {
				["copper"] = 1425000,
				["link"] = "|cff0070dd|Hitem:49908::::::::50:66:::::::|h[Urtümliches Saronit]|h|r",
				["ID"] = 49908,
			},
			[43102] = {
				["copper"] = 712500,
				["link"] = "|cff0070dd|Hitem:43102::::::::50:66:::::::|h[Gefrorene Kugel]|h|r",
				["ID"] = 43102,
			},
			[18567] = {
				["copper"] = 30000,
				["link"] = "|cffffffff|Hitem:18567::::::::50:66:::::::|h[Elementarfluxus]|h|r",
				["ID"] = 18567,
			},
			[180733] = {
				["copper"] = 50000,
				["link"] = "|cffffffff|Hitem:180733::::::::50:66:::::::|h[Leuchtender Fluxus]|h|r",
				["ID"] = 180733,
			},
	}

function Item:RegisterVendorMats(link)

	local vendorMat = {}
	vendorMat.ID = Item:GetItemID(link)

	if not tonumber(vendorMat.ID) then
		Item:DebugPrintf("ERR~RegisterVendorMats(%s): no numeric itemID", tostring(link))
		return
	end

	if private.vendorMats[vendorMat.ID] then
		Item:DebugPrintf("WARN~9~RegisterVendorMats(%s): already configured", tostring(link))
		return
	end

	local itemName, itemLink, _, _, _, itemType, itemSubType, _,
		_, _, _, itemClassID, itemSubClassID, _, _, _, isCraftingReagent = GetItemInfo(link)

	addon.db.global.vendorMats = addon.db.global.vendorMats or {}
	if addon.db.global.vendorMats[vendorMat.ID] then
		Item:DebugPrintf("WARN~9~RegisterVendorMats(%s): already registered", tostring(link))
		return
	end

	if itemName then
		vendorMat.name = itemName
		vendorMat.link = itemLink
		vendorMat.type = itemType -- Handwerksmaterial
		vendorMat.subType = itemSubType -- Sonstiges
		vendorMat.classID = itemClassID -- 7
		vendorMat.subClassID = itemSubClassID -- 11
		vendorMat.reagent = isCraftingReagent -- true
		vendorMat.copper = 0
	else
		Item:DebugPrintf("ERR~RegisterVendorMats(%s): GetItemInfo failed", tostring(link))
		return
	end

	if vendorMat.reagent then
		addon.db.global.vendorMats[vendorMat.ID] = vendorMat
		Item:DebugPrintf("8~RegisterVendorMats: ID=%s, link=%s", tostring(vendorMat.ID), tostring(vendorMat.link))
	else
		addon.db.global.vendorMats[vendorMat.ID] = nil
	end
end

function Item:RecipeID2Mats(recipeID)
	-- Item:DebugPrintf("RecipeID2Mats(%s)", tostring(recipeID))

	local matsID, matsCnt

	local numReagents = C_TradeSkillUI.GetRecipeNumReagents(recipeID)

	for reagentIndex = 1, numReagents do
		local link = C_TradeSkillUI.GetRecipeReagentItemLink(recipeID, reagentIndex)

		if reagentIndex == 1 then
			matsID = Item:GetItemID(link)
			local _, _, reagentCount = C_TradeSkillUI.GetRecipeReagentInfo(recipeID, reagentIndex)
			matsCnt = reagentCount
		end

		-- Item:DebugPrintf("  link=%s", tostring(link))
		Item:RegisterVendorMats(link)
	end
	return matsID, matsCnt
end

-- Return an upgraded item link
function Item:MakeRelicLink(link, id)
	local llink
	if id == 180055 then
		llink = gsub(link, "1:3524:", "2:3524:7175:")
	end
	if id == 180057 then
		llink = gsub(link, "1:3524:", "2:3524:7176:")
	end
	if id == 180058 then
		llink = gsub(link, "1:3524:", "2:3524:7177:")
	end
	if id == 180059 then
		llink = gsub(link, "1:3524:", "2:3524:7178:")
	end
	if id == 180060 then
		llink = gsub(link, "1:3524:", "2:3524:7179:")
	end
	if id == 183942 then
		llink = gsub(link, "1:3524:", "2:3524:7179:")
	end
	if id == 173381 then
		llink = gsub(link, "1:3524:", "2:3524:7180:")
	end
	if id == 173382 then
		llink = gsub(link, "1:3524:", "2:3524:7185:")
	end
	return llink
end

-- Return an escaped link for diagnose
function Item:EscapedLink(link)
	local l = ""
	if type(link) == "string" and link:match("^\124(.+)\124h") then
		l = gsub(link, "\124", "\124\124")
	end
	return l
end

-- Return a shortened Relic/Mark name
function Item:ReagentShort(id)
	if not id or id == "nil" or id == "" then
		return ""
	end

	local name = Item:GetItemInfo(id) or tostring(id)
	name = gsub(name, L["Relic of the Past"], L["Relic"])
	name = gsub(name, L["Novice Crafter's Mark"], L["NovMark"])
	name = gsub(name, L["Crafter's Mark"], L["Mark"])
	return name
end

-- Return itemID from itemLink
function Item:GetItemID(val)
	if type(val) == "number" then
		return val
	end

	if type(val) == "string" then
		local id = val:match("Hitem:([0-9]+)")
		if id and tonumber(id) then
			return tonumber(id)
		end
	end
end

-- Return itemName from itemLink
function Item:GetItemName(val)
	if type(val) == "string" then
		local id = val:match("|h(.+)|h")
		if id then
			return id
		end
	end
end

-- Return money string from copper
function Item:FromCopper(copper, isVS)
	if not tonumber(copper) then return end

	local g = floor(copper / COPPER_PER_GOLD)
	local s = floor((copper - (g * COPPER_PER_GOLD)) /  COPPER_PER_SILVER)
	local c = copper - (g * COPPER_PER_GOLD) - (s * COPPER_PER_SILVER)

	local result = ""

	local vsC = ""
	local vsR = ""
	if isVS then
		vsC = ITEM_QUALITY_COLORS[7].hex
		vsR = FONT_COLOR_CODE_CLOSE
	end

	if not addon.db.global.discardSilver then
		if not addon.db.global.discardCopper then
			if (s > 0) or (g > 0) then
				result = vsC .. format("%02d", c) .. vsR .. "|cffeda55fc|r"
			else
				result = vsC .. format("%d", c) .. vsR .. "|cffeda55fc|r"
			end
		end
		if addon.db.global.discardCopper then
			if g == 0 and s == 0 then
				result = vsC .. format("%d", c) .. vsR .. "|cffeda55fc|r"
			end
		end

		if g > 0 then
			result = vsC .. format("%02d", s) .. vsR .. "|cffc7c7cfs|r " .. result
		else
			if s > 0 then
				result = vsC .. format("%d", s) .. vsR .. "|cffc7c7cfs|r " .. result
			end
		end
	end
	if addon.db.global.discardSilver then
		if g == 0 and s > 0 then
			result = vsC .. format("%d", s) .. vsR .. "|cffc7c7cfs|r " .. result
		end
		if g == 0 and s == 0 then
			result = vsC .. format("%d", c) .. vsR .. "|cffeda55fc|r"
		end
	end
	if g > 0 then
		local gs = tostring(g)
		gs = gsub(gs, "(%d)(%d%d%d)$", "%1" .. LARGE_NUMBER_SEPERATOR .. "%2")
		result = vsC .. gs .. vsR .. "|cffffd70ag|r " .. result
	end

	return result
end

------------------------------------------------------------------------------
-- Retreive Recipe/Item infos and fill recipe list log

function Item:SellsAddRecipe(this, list, recipeID, name)
	Item:DebugPrintf("8~SellsAddRecipe for %s: %s (%s)", tostring(list), tostring(name), tostring(recipeID))

	local itemLink = C_TradeSkillUI.GetRecipeItemLink(recipeID)
	local recipeInfo = C_TradeSkillUI.GetRecipeInfo(recipeID)
	local numAvailable = recipeInfo.numAvailable or 0
	numAvailable = numAvailable > 0 and numAvailable or nil
	local itemName, _, vendorSell, baseLvl, effLvl = Item:GetItemInfo(itemLink)

	-- fill table with vendor mats
	local matsID, matsCnt
	local matsSell = 0

	if addon.db.global.isExp then
		matsID, matsCnt = Item:RecipeID2Mats(recipeID)

		if vendorSell then
			if tonumber(matsCnt) and matsCnt > 0 then
				matsSell = floor(vendorSell / matsCnt)
			end
		end
	end

	-- 1. current recipeID
	addon.GUI.DebugLog(list, "1=%s# 2=%s# 3=%s# 4=%s# 5=%s# 6=%s# 7=%s# 8=%s# 9=%s# 10=%s# 11=%s#",
		tostring(recipeID), tostring(numAvailable), "---", tostring(itemName), tostring(itemLink),
		tostring(vendorSell), tostring(baseLvl), tostring(effLvl),
		tostring(matsID), tostring(matsCnt), tostring(matsSell))

	-- 2. add all variants of items made with optional reagents
	if C_TradeSkillUI.GetOptionalReagentInfo and C_TradeSkillUI.GetOptionalReagentInfo(recipeID)
		and C_TradeSkillUI.GetOptionalReagentInfo(recipeID)[1] then

		for _, v in ipairs(C_TradeSkillUI.GetOptionalReagentInfo(recipeID)) do
			local optionalReagentSlots = v.options
			for _, id in ipairs(optionalReagentSlots) do
				local lLink = Item:MakeRelicLink(itemLink, id)
				if lLink then
					local cName, cLink, cVendorSell, cBaseLvl, cEffLvl = Item:GetItemInfo(lLink)

					local cMatsSell = 0
					-- if cVendorSell then
						-- relict shuffle is dead!
						-- if tonumber(matsCnt) and matsCnt > 0 then
						-- 	cMatsSell = floor(cVendorSell / matsCnt)
						-- end
					-- end

					addon.GUI.DebugLog(list, "1=%s# 2=%s# 3=%s# 4=%s# 5=%s# 6=%s# 7=%s# 8=%s# 9=%s# 10=%s# 11=%s#",
						tostring(recipeID), tostring(numAvailable), Item:ReagentShort(tostring(id)), tostring(cName), tostring(cLink),
						tostring(cVendorSell), tostring(cBaseLvl), tostring(cEffLvl),
						tostring(matsID), tostring(matsCnt), tostring(cMatsSell))
				end
			end
		end
	end
end

------------------------------------------------------------------------------
-- Sets self disappearing status text
function Item:SetStatusLine(text, seconds)
	private.statusLine = text
	private.statusSeconds = seconds
end

------------------------------------------------------------------------------
-- GetItemInfo Cache + Delayed Update

private.gotNIL = 0
private.gotNILMaxTry = 5
private.gotNILTry = 5

function Item:ResetNILTry()
	Item:DebugPrintf("ResetNILTry()")
	private.gotNILTry = 5
end

function Item:SecTimer()
	-- Item:DebugPrintf("SecTimer()")

	-- ItemInfoCache
	if private.gotNIL > 0 then
		Item:DebugPrintf("ERR~SecTimer: gotNIL=%s gotNILTry=%s", tostring(private.gotNIL), tostring(private.gotNILTry))
		Item:SetStatusLine(L["Refreshing cache for GetItemInfo()..."], 2)
		private.gotNIL = 0
		if private.gotNILTry > 0 and addon.GUI.display and TradeSkillFrame and TradeSkillFrame.RecipeList and TradeSkillFrame:IsShown() then
			private.gotNILTry = private.gotNILTry - 1
			if addon.GUI.display and addon.GUI.view then
				Item:DebugPrintf("ERR~  -> Hooks:FillLogFromRecipes/DebugLog_UpdateData(%s)", tostring(addon.GUI.view))
				addon.Hooks:FillLogFromRecipes(addon.GUI.view)
				addon.GUI.DebugLog_UpdateData(addon.GUI.view, true)
			end
		end
		-- just to be sure
		if not addon.GUI.display then
			Item:DebugPrintf("ERR~SecTimer: gotNIL reset")
			private.gotNIL = 0
			private.gotNILTry = private.gotNILMaxTry
		end
	end

	-- Statusline
	if private.statusLine and strlen(private.statusLine) > 0 then
		private.statusSeconds = private.statusSeconds or 1
		if private.statusFlag ~= nil and private.statusFlag >= private.statusSeconds then
			private.statusFlag = 0
			private.statusLine = ""
		else
			private.statusFlag = private.statusFlag or 0
			private.statusFlag = private.statusFlag + 1
		end
		addon.GUI:SetStatusLine(private.statusLine)
	end
end

function Item:GetItemInfo(link)
	if not link or type(link) ~= "string" then
		return
	end

	private.itemInfoCache = private.itemInfoCache or {}

	if link:match("Henchant") then
		Item:DebugPrintf("8~GetItemInfo(%s/%s): enchant", tostring(link), tostring(Item:EscapedLink(link)))
		private.itemInfoCache[link] = {name = nil, link = nil, vendorSell = nil, baseLvl = nil, effLvl = nil}
		return
	end

	if private.itemInfoCache[link] then
		-- Item:DebugPrintf("GetItemInfo(%s): hit", tostring(link))
		return private.itemInfoCache[link].name, private.itemInfoCache[link].link,
			private.itemInfoCache[link].vendorSell, private.itemInfoCache[link].baseLvl, private.itemInfoCache[link].effLvl
	else
		local cName, cLink, _, cLevel, _, _, _, _, _, _, cVendorSell = GetItemInfo(link)
		local eLevel = GetDetailedItemLevelInfo(link)
		if cName and eLevel then
			Item:DebugPrintf("8~GetItemInfo(%s/%s/%s): no hit", tostring(link), tostring(Item:EscapedLink(link)), tostring(eLevel))
			private.itemInfoCache[link] = {name = cName, link = cLink, vendorSell = cVendorSell, baseLvl = cLevel, effLvl = eLevel}
			return private.itemInfoCache[link].name, private.itemInfoCache[link].link,
				private.itemInfoCache[link].vendorSell, private.itemInfoCache[link].baseLvl, private.itemInfoCache[link].effLvl
		else
			Item:DebugPrintf("ERR~7~GetItemInfo(%s/%s): no data", tostring(link), tostring(Item:EscapedLink(link)))
			private.gotNIL = private.gotNIL + 1
			Item:DebugPrintf("ERR~3~Set gotNIL=%s", tostring(private.gotNIL))
		end
	end
end

function Item:GetItemValue(link, source)
	if not link or type(link) ~= "string" then
		return
	end

	if not source or type(source) ~= "string" then
		return
	end

	local key = tostring(link) .. tostring(source)

	private.itemValueCache = private.itemValueCache or {}

	if link:match("Henchant") then
		Item:DebugPrintf("8~GetItemValue(%s/%s): enchant", tostring(link), tostring(Item:EscapedLink(link)))
		private.itemValueCache[key] = {}
		private.itemValueCache[key] = nil
		return
	end

	if private.itemValueCache[key] then
		-- Item:DebugPrintf("GetItemValue(%s): hit", tostring(key))
		return private.itemValueCache[key]
	else
		if source:match("^TSM:(.+)$") then
			local itemString
			if TSM_API and TSM_API.ToItemString then
				local status, res = pcall(TSM_API.ToItemString, link)
				if status then
					itemString = res
				else
					Item:DebugPrintf("ERR~  broken TSM4 API: pcall of TSM_API.ToItemString=%s", tostring(res))
				end
			end
			if itemString and TSM_API and TSM_API.GetCustomPriceValue and TSM_API.ToItemString then
				local status, res = pcall(TSM_API.GetCustomPriceValue, strlower(source:match("^TSM:(.+)$")), itemString)
				if status then
					Item:DebugPrintf("  = %s", tostring(res))
					private.itemValueCache[key] = res
				else
					Item:DebugPrintf("ERR~  broken TSM4 API: pcall of TSM_API.GetCustomPriceValue=%s", tostring(res))
				end
			end
		end

		if private.itemValueCache[key] then
			Item:DebugPrintf("8~GetItemValue(%s/%s/%s): no hit", tostring(link), tostring(Item:EscapedLink(link)), tostring(private.itemValueCache[key]))
			return private.itemValueCache[key]
		else
			Item:DebugPrintf("ERR~7~GetItemValue(%s/%s): no data", tostring(link), tostring(Item:EscapedLink(link)))
			return nil
		end
	end
end

-- EOF
